module lexer;
import std.string;
import std.ctype;
import std.conv;

//-----------------------------------------------------------------------
// Token
//-----------------------------------------------------------------------

bool VerboseToString = false;

enum
{
	EOP,
	LPAREN,
	RPAREN,
	COMMA,
	DOT,
	CMINUS,
	QMINUS,
	WILD,
	VAR,
	CONST,
}

typedef int SymId;

class Token
{
	int    type;
	SymId  id   = -1;  // for VAR and CONST
	string name = "@"; // for VAR and CONST

	string toString()
	{
		switch( type )
		{
		case .EOP:    return "[EOP]";
		case .LPAREN: return "(";
		case .RPAREN: return ")";
		case .COMMA:  return ",";
		case .DOT:    return ".";
		case .CMINUS: return ":-";
		case .QMINUS: return "?-";
		case .WILD:   return "_";
		case .VAR:    return VerboseToString ? "<" ~name~ ":" ~to!(string)(cast(int)id)~ ">" : name;
		case .CONST:  return VerboseToString ? "<" ~name~ ":" ~to!(string)(cast(int)id)~ ">" : name;
		default:      return "@BUG:UNKNOWNTOKEN@";
		}
	}

private: // only the lexer can construct Tokens
	static this()
	{
		EOP    = new Token(.EOP);
		LPAREN = new Token(.LPAREN);
		RPAREN = new Token(.RPAREN);
		COMMA  = new Token(.COMMA);
		DOT    = new Token(.DOT);
		CMINUS = new Token(.CMINUS);
		QMINUS = new Token(.QMINUS);
		WILD   = new Token(.WILD);
	}

	static Token EOP;
	static Token LPAREN;
	static Token RPAREN;
	static Token COMMA;
	static Token DOT;
	static Token CMINUS;
	static Token QMINUS;
	static Token WILD;
	static Token CONST( ref string[SymId] i2v, ref SymId[string] v2i, string v )
		{ return new Token(.CONST, newId(i2v, v2i, v), v); }
	static Token VAR( ref string[SymId] i2v, ref SymId[string] v2i, string v )
		{ return new Token(.VAR, newId(i2v, v2i, v), v); }

	this( int t ) { this.type = t; }
	this( int t, SymId i, string n ) { this.type=t; this.id=i; this.name=n; }
}

//-----------------------------------------------------------------------
// Util
//-----------------------------------------------------------------------

private V newId(K,V)( ref K[V] v2k, ref V[K] k2v, K k )
{
	if( k in k2v )
		return k2v[k];
	V v = cast(V) k2v.length;
	k2v[k] = v;
	v2k[v] = k;
	return v;
}

private void skipws( ref string heystack, ref int lineCounter )
{
	// munch( heystack, " \t\r\n" ); // with lineNo counting
	while( heystack.length>0
		&& (heystack[0]==' '  || heystack[0]=='\n'
		 || heystack[0]=='\t' || heystack[0]=='\r') )
	{
		if( heystack[0]=='\n' )
			lineCounter++;
		heystack = heystack[1..$];
	}
}

private string eat( ref string heystack, string needle )
{
	if( needle.length<=heystack.length
	 && heystack[0..needle.length] == needle )
	{
		heystack = heystack[needle.length..$];
		return needle;
	}
	return null;
}

private string eatAlnum( ref string heystack )
{
	return munch(heystack, "a-zA-Z0-9_");
}

//-----------------------------------------------------------------------
// Lexer
//-----------------------------------------------------------------------

class Lexer
{
	this( string src_text )
	{
		this.source_left = src_text;
	}

	Token peek()
	{
		auto baks = source_left;
		auto bakl = lineNo;
		scope(exit) {source_left = baks; lineNo = bakl;}
		return eat();
	}

	Token eat()
	{
		source_left.skipws(lineNo);

		if( source_left.length == 0 )      return Token.EOP;
		else if( islower(source_left[0]) ) return Token.CONST(id2k, k2id, source_left.eatAlnum());
		else if( isupper(source_left[0]) ) return Token.VAR(id2v, v2id, source_left.eatAlnum());
		else if( source_left.eat("(") )    return Token.LPAREN;
		else if( source_left.eat(")") )    return Token.RPAREN;
		else if( source_left.eat(",") )    return Token.COMMA;
		else if( source_left.eat(".") )    return Token.DOT;
		else if( source_left.eat("_") )    return Token.WILD;
		else if( source_left.eat(":-") )   return Token.CMINUS;
		else if( source_left.eat("?-") )   return Token.QMINUS;

		throw new Exception("Unknown Token @ " ~ toString());
	}

	string toString()
	{
		if( source_left.length == 0 )
			return "[[Lexer End]]";

		int len = 5 < source_left.length ? 5 : source_left.length;
		string context = source_left[0..len].replace("\n", "     ");
		return "[[Lexer L." ~to!(string)(lineNo)~ " : " ~context~ "]]";
	}

private:
	// reader state
	string source_left;
	SymId  lineNo = 1;

	// fresh id state
	string[SymId] id2v; SymId[string] v2id;
	string[SymId] id2k; SymId[string] k2id;
}

//-----------------------------------------------------------------------
// Test
//-----------------------------------------------------------------------

unittest
{
	auto lex = new Lexer(``);
	assert( lex.eat() is Token.EOP );
}

unittest
{
	auto lex = new Lexer(`hello(World).`);
	assert( lex.eat().id == 0 );
	assert( lex.eat() is Token.LPAREN );
	assert( lex.eat().id == 0 );
	assert( lex.eat() is Token.RPAREN );
	assert( lex.eat() is Token.DOT );
	assert( lex.eat() is Token.EOP );
}

unittest
{
	auto lex = new Lexer(`hello :- bar(Foo), Foo. hello(Buz,Buz).`);
	assert( lex.eat().id == 0 );
	assert( lex.eat() is Token.CMINUS );
	assert( lex.eat().id == 1 );
	assert( lex.eat() is Token.LPAREN );
	assert( lex.eat().id == 0 );
	assert( lex.eat() is Token.RPAREN );
	assert( lex.eat() is Token.COMMA );
	assert( lex.eat().id == 0 );
	assert( lex.eat() is Token.DOT );
	assert( lex.eat().id == 0 );
	assert( lex.eat() is Token.LPAREN );
	assert( lex.eat().id == 1 );
	assert( lex.eat() is Token.COMMA );
	assert( lex.eat().id == 1 );
	assert( lex.eat() is Token.RPAREN );
	assert( lex.eat() is Token.DOT );
	assert( lex.eat() is Token.EOP );
}
