module parser;
import lexer;
import std.conv;

class Term
{
	Token  t;
	Term[] child;

	string toString()
	{
		string str = t.toString();
		if( child.length>0 ) {
			str ~= "(";
			foreach(i,t; child) {
				if(i) str ~= ", ";
				str ~= t.toString();
			}
			str ~= ")";
		}
		return str;
	}
}

class Query
{
	Term[] tail;

	string toString()
	{
		string str = "?- ";
		foreach(i,t; tail) {
			if(i) str ~= ", ";
			str ~= t.toString();
		}
		str ~= ".";
		return str;
	}
}

class Decl
{
	Term   head;
	Term[] tail;

	string toString()
	{
		string str = head.toString() ~ " :- ";
		foreach(i,t; tail) {
			if(i) str ~= ", ";
			str ~= t.toString();
		}
		str ~= ".";
		return str;
	}
}

class Program
{
	Decl[]   decl;
	Query[]  query;

	int[int] arity; // funsym -> arity
}


class Parser
{
	this( Lexer lex ) { this.lex = lex; }
	private Lexer lex;

	Program parse_program()
	{
		Program prog = new Program;
		while( lex.peek().type != EOP )
			if( lex.peek().type == QMINUS )
				prog.query ~= parse_query( prog.arity );
			else
				prog.decl  ~= parse_decl( prog.arity );
		return prog;
	}

	Term parse_term( ref int[int] arity )
	{
		Term term = new Term;

		Token t = lex.eat();
		if( t.type==CONST || t.type==VAR || t.type==WILD )
		{
			term.t = t;
			bool fst = true;
			if( lex.peek().type == LPAREN )
			{
				if( t.type != CONST )
					throw new Exception("Variables cannot have parameters @ "~lex.toString());

				for(;;) {
					Token tt = lex.eat();
					if( fst ) {
						if( tt.type != LPAREN )
							throw new Exception("LPAREN required @ "~lex.toString());
						fst = false;
						term.child ~= parse_term(arity);
					} else {
						if( tt.type == RPAREN )
							break;
						else if( tt.type == COMMA )
							{ term.child ~= parse_term(arity); }
						else
							throw new Exception("COMMA or RPAREN required @ "~lex.toString());
					}
				}
			}
		}
		else
			throw new Exception("CONST or VAR required @ "~lex.toString());

		// arity check
		if( term.t.type == CONST )
		{
			if( term.t.id in arity && arity[term.t.id]!=term.child.length )
				throw new Exception("Arity mismatch: "~to!(string)(term.t)
					~" "~to!(string)(term.child.length)~" vs "
					~to!(string)(arity[term.t.id])~" @ "~lex.toString());
			arity[term.t.id] = term.child.length;
		}
		return term;
	}

	Query parse_query( ref int[int] arity )
	{
		if( lex.peek().type == EOP )
			return null;

		Query query = new Query;
		Token t = lex.eat();
		if( t.type == QMINUS )
			query.tail ~= parse_terms_dot( arity );
		else
			throw new Exception("QMINUS required @ "~lex.toString());
		return query;
	}

	private Term[] parse_terms_dot( ref int[int] arity )
	{
		Term[] lst;
		for(;;)
		{
			lst ~= parse_term( arity );
			Token tt = lex.eat();
			if( tt.type == DOT )
				break;
			else if( tt.type == COMMA )
				continue;
			else
				throw new Exception("DOT or COMMA required @ "~lex.toString());
		}
		return lst;
	}

	Decl parse_decl( ref int[int] arity )
	{
		if( lex.peek().type == EOP || lex.peek().type==QMINUS )
			return null;

		Decl decl = new Decl;
		decl.head = parse_term( arity );

		Token t = lex.eat();
		if( t.type == CMINUS ) // rule
			decl.tail ~= parse_terms_dot( arity );
		else if( t.type == DOT ) // fact
			{}
		else
			throw new Exception("CMINUS or DOT required @ "~lex.toString());

		return decl;
	}
}
