module runtime;
import lexer;
import parser;
import std.conv;
import std.stdio;

//-----------------------------------------------------------------------

abstract class VM_Term
{
}

class VM_Var  : VM_Term
{
	VM_Term t; // null if not unified

	string toString()
	{
		if( t !is null )
			return t.toString();
		return "<$:"~to!(string)(toHash)~">";
	}
}

class VM_Node : VM_Term
{
	Token     t;
	VM_Term[] child;
	this( typeof(this.tupleof) xs ) {this.tupleof = xs;}

	string toString()
	{
		string s = t.toString();
		if( child.length ) {
			s ~= "(";
			foreach(i,c; child) {
				if(i) s ~= ", ";
				s ~= c.toString();
			}
			s ~= ")";
		}
		return s;
	}
}

VM_Term[] instanciate( Decl d )
{
	return instanciate(d.head ~ d.tail);
}

VM_Term[] instanciate( Term[] ts )
{
	VM_Var[SymId] assigned;
	return instanciate(ts, assigned);
}

VM_Term[] instanciate( Term[] ts, ref VM_Var[SymId] assigned )
{
	VM_Term[] rts;
	foreach(t; ts)
		rts ~= instanciate(t, assigned);
	return rts;
}

VM_Term instanciate( Term ts, ref VM_Var[SymId] assigned )
{
	if( ts.t.type == WILD )
	{
		return new VM_Var;
	}
	else if( ts.t.type == VAR )
	{
		if( !(ts.t.id in assigned) )
			assigned[ts.t.id] = new VM_Var;
		return assigned[ts.t.id];
	}
	else
	{
		return new VM_Node(ts.t, instanciate(ts.child, assigned));
	}
}

//-----------------------------------------------------------------------
// VM instructions
//-----------------------------------------------------------------------

abstract class Code {}

alias Code[] CodeBlock;

class OP_LoadHorn : Code
{
	Decl d;
	this( typeof(this.tupleof) xs ) {this.tupleof = xs;}
}
class OP_LoadGoal : Code
{
	Query q;
	this( typeof(this.tupleof) xs ) {this.tupleof = xs;}
}
class OP_BTPoint : Code
{
	CodeBlock next;
	this( typeof(this.tupleof) xs ) {this.tupleof = xs;}
}
class OP_UnifyHead : Code
{
}
class OP_ExpandGoal : Code
{
}
class OP_MainLoop : Code
{
}
class OP_Fail : Code
{
}
class OP_End : Code
{
}

//-----------------------------------------------------------------------
// Compile
//-----------------------------------------------------------------------

CodeBlock compile( Decl[] decl )
{
	if( decl.length == 0 )
		return [new OP_Fail];
	else
		return compile( decl[0], compile(decl[1..$]) );
}

CodeBlock compile( Decl d, CodeBlock next )
{
	CodeBlock cb;
	cb ~= new OP_LoadHorn(d);
	cb ~= new OP_BTPoint(next);
	cb ~= new OP_UnifyHead();
	cb ~= new OP_ExpandGoal();
	cb ~= new OP_MainLoop();
	return cb;
}

class QueryBlock
{
	Query orig;
	CodeBlock cb;
	VM_Term[] goal;

	Token[]   fv;
	VM_Var[]  fvVar;
}

Token[] freeVar( Term[] ts, ref bool[SymId] vis )
{
	Token[] result;
	foreach(t; ts)
		result ~= freeVar(t, vis);
	return result;
}
Token[] freeVar( Term ts, ref bool[SymId] vis )
{
	Token[] result;
	if( ts.t.type == CONST ) {
		foreach(t; ts.child)
			result ~= freeVar(t, vis);
	} else if( ts.t.type == VAR ) {
		if( !(ts.t.id in vis) ) {
			vis[ts.t.id] = true;
			return [ts.t];
		}
		return [];
	}
	return result;
}

QueryBlock compile( Query q )
{
	CodeBlock cb;
	cb ~= new OP_LoadGoal(q);
	cb ~= new OP_BTPoint( [new OP_End] );
	cb ~= new OP_MainLoop();

	QueryBlock qb = new QueryBlock;
	qb.orig = q;
	qb.cb = cb;
	bool[SymId] vis;
	qb.fv = freeVar(q.tail, vis);
	VM_Var[SymId] assigned;
	qb.goal = instanciate(q.tail, assigned);
	foreach(fv; qb.fv)
		qb.fvVar ~= assigned[fv.id];
	return qb;
}

//-----------------------------------------------------------------------
// Backtracking Information
//-----------------------------------------------------------------------

class BackSteppo
{
	VM_Term[]  goal;
	VM_Term[]  horn;
	CodeBlock  next;
	VM_Term*[] modified;
	this( typeof(this.tupleof) xs ) {this.tupleof = xs;}

	void recover()
	{
		foreach(p; modified)
			*p = null;
	}

	void update( VM_Term* p, VM_Term t )
	{
		*p = t;
		modified ~= p;
	}

	bool unify( VM_Term a, VM_Term b )
	{
		while( cast(VM_Var)a && (cast(VM_Var)a).t ) a = (cast(VM_Var)a).t;
		while( cast(VM_Var)b && (cast(VM_Var)b).t ) b = (cast(VM_Var)b).t;

		if( cast(VM_Node)a && cast(VM_Node)b )
		{
			if( (cast(VM_Node)a).t.id == (cast(VM_Node)b).t.id )
			{
				foreach(i,_; (cast(VM_Node)a).child )
					if( !unify( (cast(VM_Node)a).child[i], (cast(VM_Node)b).child[i] ) )
						return false;
			}
			else
				return false;
		}
		else if( cast(VM_Var)a )
		{
			update( &(cast(VM_Var)a).t, b );
		}
		else // if( cast(VM_Var)b )
		{
			update( &(cast(VM_Var)b).t, a );
		}
		return true;
	}
}

//-----------------------------------------------------------------------
// VM Interface
//-----------------------------------------------------------------------

class Runtime
{
	this( Program p )
	{
		main = compile( p.decl );
		foreach(q; p.query)
			query ~= compile(q);
	}

	int length()
	{
		return query.length;
	}

	void run_query( int i )
	{
		vm_loop(query[i], main);
	}

private:
	CodeBlock    main;
	QueryBlock[] query;

private:
	void vm_loop( QueryBlock query, CodeBlock main )
	{
		writeln("QUERY ", query.orig);

		// machine state
		VM_Term[]    goal;
		VM_Term[]    horn;
		BackSteppo[] bstk;
		CodeBlock    cb = query.cb;

		// backtrack operation
		void backtrack()
		{
			goal = bstk[$-1].goal;
			horn = bstk[$-1].horn;
			cb   = bstk[$-1].next;
			bstk[$-1].recover();
			bstk = bstk[0..$-1];
		}

		int nAns = 0;

		void found()
		{
			++nAns;
			writeln("OK");
			foreach(i,_; query.fv)
			{
				auto v = query.fv[i];
				auto t = query.fvVar[i];
					writeln("  ", v, " = ", t);
			}
		}

		void end()
		{
			writeln("END - ", nAns, " answers.");
			writeln();
		}

		for(;;)
		{
			Code c = cb[0];
			cb = cb[1..$];

			if( auto e = cast(OP_LoadHorn)c )
			{
				horn = instanciate(e.d);
			}
			else if( auto e = cast(OP_LoadGoal)c )
			{
				goal = query.goal;
			}
			else if( auto e = cast(OP_BTPoint)c )
			{
				bstk ~= new BackSteppo(goal, horn, e.next, null);
			}
			else if( auto e = cast(OP_UnifyHead)c )
			{
				if( !bstk[$-1].unify( goal[0], horn[0] ) )
					backtrack();
			}
			else if( auto e = cast(OP_ExpandGoal)c )
			{
				goal = horn[1..$] ~ goal[1..$];
				if( goal.length == 0 )
				{
					found();
					backtrack();
				}
			}
			else if( auto e = cast(OP_MainLoop)c )
			{
				cb = main;
			}
			else if( auto e = cast(OP_Fail)c )
			{
				backtrack();
			}
			else if( auto e = cast(OP_End)c )
			{
				end();
				break;
			}
		}
	}
}
